var self               = require('sdk/self'),
    data               = self.data,
    sp                 = require('sdk/simple-prefs'),
    buttons            = require('sdk/ui/button/action'),
    privateBrowsing    = require("sdk/private-browsing"),
    tabs               = require('sdk/tabs'),
    timers             = require('sdk/timers'),
    loader             = require('@loader/options'),
    unload             = require('sdk/system/unload'),
    pageMod            = require("sdk/page-mod"),
    tUtils             = require('sdk/tabs/utils'),
    array              = require('sdk/util/array'),
    {Cc, Ci, Cu}       = require('chrome'),
    prefs              = sp.prefs,
    config             = require("../config");

exports.timer = timers;
exports.loadReason = self.loadReason;
var {modelFor} = require("sdk/model/core");
exports.version = function () {return self.version};
var {WebRequest} = Cu.import("resource://gre/modules/WebRequest.jsm");
var {MatchPattern} = Cu.import("resource://gre/modules/MatchPattern.jsm");
exports.XMLHttpRequest = function () {return Cc['@mozilla.org/xmlextras/xmlhttprequest;1'].createInstance(Ci.nsIXMLHttpRequest)};

exports.toHostname = function (url) {
  var URL = require("sdk/url").URL;
  var HOST = (new URL(url).hostname) || url;
  return HOST;
};

var popup = require("sdk/panel").Panel({
  "width": 400,
  "height": 190,
  "contentURL": data.url("./popup/popup.html"),
  "contentScriptFile": [data.url("popup/firefox/firefox.js"), data.url("popup/popup.js")]
});

popup.on('show', function() {popup.port.emit('show', true)});

exports.popup = {
  "send": function (id, data) {popup.port.emit(id, data)},
  "receive": function (id, callback) {popup.port.on(id, callback)}
};

exports.button = (function () {
  var onClick;
  var button = buttons.ActionButton({
    "id": self.name,
    "label": 'Smart HTTPS',
    "onClick": function () {popup.show({"position": button})},
    "icon": {
      '16': './icons/enabled/16.png',
      '32': './icons/enabled/32.png',
      '64': './icons/enabled/64.png'
    }
  });
  /*  */
  return {
    set icon (obj) {button.icon = obj},
    set label (label) {button.label = label}
  };
})();

exports.storage = {
  "read": function (id) {return (prefs[id] || prefs[id] + '' === 'false' || !isNaN(prefs[id])) ? (prefs[id] + '') : null},
  "write": function (id, data) {
    data = data + '';
    if (data === 'true' || data === 'false') prefs[id] = data === 'true' ? true : false;
    else if (parseInt(data) + '' === data) prefs[id] = parseInt(data);
    else prefs[id] = data + '';
  }
};

exports.tabs = {
  "open": function (url) {tabs.open({"url": url, "inBackground": false})},
  "getActive": function (callback) {try {callback(tabs.activeTab)} catch (e) {}},
  "update": function (tab, url, flag) {flag ? tab.url = url : tUtils.setTabURL(tab, url)},
  "openOptions": function () {
    var optionsTab = false;
    for each (var tab in tabs) {
      if (tab.url === data.url('options/options.html')) {
        tab.reload();
        tab.activate();
        optionsTab = true;
        tab.window.activate();
      }
    }
    if (!optionsTab) tabs.open({url: data.url('options/options.html'), inBackground: false});
  }
};

sp.on("openOptions", function() {
  tabs.open({"url": data.url("options/options.html"), "inBackground": false});
});

exports.options = (function () {
  var workers = [], options_arr = [];
  pageMod.PageMod({
    "contentScriptWhen": "start",
    "include": data.url("options/options.html"),
    "contentScriptOptions": {"base": loader.prefixURI},
    "contentScriptFile": [data.url("options/firefox/firefox.js"), data.url("options/options.js")],
    "onAttach": function (worker) {
      array.add(workers, worker);
      worker.on('pageshow', function () {array.add(workers, this)});
      worker.on('detach', function () {array.remove(workers, this)});
      worker.on('pagehide', function () {array.remove(workers, this)});
      options_arr.forEach(function (arr) {worker.port.on(arr[0], arr[1])});
    }
  });
  /*  */
  return {
    "receive": function (id, callback) {options_arr.push([id, callback])},
    "send": function (id, data) {
      workers.forEach(function (worker) {
        if (!worker || !worker.url) return;
        worker.port.emit(id, data);
      });
    }
  }
})();

exports.tabQuery = function (details, callback) {
  if (details && details.browser) {
    try {
      var tab = tUtils.getTabForBrowser(details.browser);
      if (prefs["incognito"]) {
        var _tab = modelFor(tab);
        var incognito = privateBrowsing.isPrivate(_tab);
        tab["incognito"] = incognito;
      }
      if (tab) callback(tab); else callback(null);
    } catch (e) {callback(null)}
  } else callback(null);
};

var onBeforeRequest, onHeadersReceived, onCompleted;

exports.onBeforeRequest = function (callback) {
  onBeforeRequest = function (details) {
    if (details.type === 'main_frame') {
      return callback(details);
    }
  };
  /*  */
  unload.when(function () {WebRequest.onBeforeRequest.removeListener(onBeforeRequest)});
};

exports.onHeadersReceived = function (callback) {
  var _domain = {};
  onHeadersReceived = function (details) {
    var id = details.windowId;
    if (details.type === 'main_frame') {
      _domain[id] = exports.toHostname(details.url);
    }
    return callback(_domain[id], details);
  };
  /*  */
  unload.when(function () {WebRequest.onHeadersReceived.removeListener(onHeadersReceived)});
};

exports.onCompleted = function (callback) {
  var _error = {}, _complete = {};
  onCompleted = function (details) {
    if (details.type === 'main_frame') {
      _complete[details.url] = true;
      _error[details.url] = details.error ? true : false;
      if (!config.addon.typemissmatch) return callback(details);
    }
  };
  /*  */
  unload.when(function () {WebRequest.onCompleted.removeListener(onCompleted)});
  /*  */
  tabs.on('load', function (tab) {
    if (config.addon.state === "enabled") {
      var _url = tab.url;
      if (_complete[_url]) {
        return callback({"url": _url, "error": _error[_url]});
      }
    }
  });
};

exports.webRequest = {
  "removeListener": {
    "onCompleted": function () {WebRequest.onCompleted.removeListener(onCompleted)},
    "onBeforeRequest": function () {WebRequest.onBeforeRequest.removeListener(onBeforeRequest)},
    "onHeadersReceived": function () {WebRequest.onHeadersReceived.removeListener(onHeadersReceived)}
  },
  "addListener": {
    "onCompleted": function () {WebRequest.onCompleted.addListener(onCompleted, {"urls": (new MatchPattern(["http://*/*", "https://*/*"]))})},
    "onBeforeRequest": function () {WebRequest.onBeforeRequest.addListener(onBeforeRequest, {"urls": (new MatchPattern(["http://*/*"]))}, ["blocking"])},
    "onHeadersReceived": function () {WebRequest.onHeadersReceived.addListener(onHeadersReceived, {"urls": (new MatchPattern(["http://*/*", "https://*/*"]))})}
  },
  "init": function () {
    exports.webRequest.removeListener.onCompleted();
    exports.webRequest.removeListener.onBeforeRequest();
    exports.webRequest.removeListener.onHeadersReceived(); 
    /*  */
    if (config.addon.state === "enabled") {
      exports.webRequest.addListener.onCompleted();
      exports.webRequest.addListener.onBeforeRequest();
      if (config.addon.typemissmatch) exports.webRequest.addListener.onHeadersReceived();
    }
  }
};

unload.when(function (reason) {
  if (reason === "uninstall") {
    var URL = config.welcome.url + "?v=" + self.version + "&type=uninstall";
    tabs.open({url: URL, inBackground: false});
  }
  /*  */
  for (let tab of tabs) {
    if (tab.url === data.url("options/options.html")) tab.close();
  }
});